from dateutil.parser import parse
import re


def str_to_date(string):
    try:
        return parse(string)
    except Exception:
        return None


# Simple url check, though very less chance of working
def _extract_from_url(url):
    regex = r'([\./\-_]{0,1}(19|20)\d{2})[\./\-_]{0,1}(([0-3]{0,1}[0-9][\./\-_])|(\w{3,5}[\./\-_]))([0-3]{0,1}[0-9][\./\-]{0,1})?' # noqa
    m = re.search(regex, url)
    if m:
        return str_to_date(m.group(0))
    return None


# Some common meta names and properties that may denote date
DATE_META_NAMES = [
    'date',
    'pubdate',
    'publishdate',
    'timestamp',
    'dc.date.issued',
    'sailthru.date',
    'article.published',
    'published-date',
    'article.created',
    'article_date_original',
    'cxenseparse:recs:publishtime',
    'date_published',
    'datepublished',
    'datecreated',
    'article:published_time',
    'bt:pubdate',
]


def _extract_from_meta(page):
    meta_date = None
    for meta in page.findAll('meta'):
        meta_name = meta.get('name', '').lower()
        item_prop = meta.get('itemprop', '').lower()
        meta_property = meta.get('property', '').lower()
        http_equiv = meta.get('http-equiv', '').lower()

        if (
                meta_name in DATE_META_NAMES or
                item_prop in DATE_META_NAMES or
                meta_property in DATE_META_NAMES or
                http_equiv == 'date'
        ):
            meta_date = str_to_date(meta['content'].strip())
            break

    return meta_date


# From https://github.com/Webhose/article-date-extractor
# Most probably can be optimized
def _extract_from_tags(page):
    for time in page.findAll('time'):
        datetime = time.get('datetime', '')
        if len(datetime) > 0:
            return str_to_date(datetime)

        datetime = time.get('class', '')
        if len(datetime) > 0 and datetime[0].lower() == 'timestamp':
            return str_to_date(time.string)

    tag = page.find('span', {'itemprop': 'datePublished'})
    if tag is not None:
        date_text = tag.get('content')
        if date_text is None:
            date_text = tag.text

        if date_text:
            return str_to_date(date_text)

    regex = 'pubdate|timestamp|article_date|articledate|date'
    for tag in page.find_all(['span', 'p', 'div'],
                             class_=re.compile(regex, re.IGNORECASE)):
        date_text = tag.string
        if date_text is None:
            date_text = tag.text

        possible_date = str_to_date(date_text)
        if possible_date:
            return possible_date

    return None


def extract_date(url, page):
    """
    Extract date from given url and given page data
    where the page is a BeautifulSoup object
    """
    if not page:
        date = _extract_from_url(url)
    else:
        date = _extract_from_meta(page)
        if date is None:
            date = _extract_from_tags(page)
        if date is None:
            date = _extract_from_url(url)

    return date and date.date()



import datetime
import os
import logging
import sentry_sdk

from .django_utils.common import DjangoJSONEncoder
from .validators import ValidationError
from .camelize import camelize_response, underscoreize

from lambda_decorators import (
    LambdaDecorator as OLambdaDecorator,
    json_http_resp,
    load_json_body,
    cors_headers,
)

logger = logging.getLogger(__name__)

CORS_DOMAIN = os.environ.get('CORS_DOMAIN', '*')
VALIDATION_EXCEPTIONS = (
    ValidationError,
)


def create_error_response(exception, event, context) -> dict:
    """Generate response from exception"""
    resp_data = {'timestamp': datetime.datetime.now().timestamp()}
    status_code = 500
    # TODO: Hanlde exception default message if present
    exc_data = exception.args[0] if len(exception.args) else None

    non_field_errors = ['Something unexpected happened']
    internal_non_field_errors = []

    # 400 response
    if exception.__class__ in VALIDATION_EXCEPTIONS:
        status_code = 400
        non_field_errors = []
        if isinstance(exc_data, dict):
            resp_data['errors'] = exc_data
        elif exc_data.__class__ in [list, tuple]:
            non_field_errors = exc_data
        elif isinstance(exc_data, str):
            non_field_errors = [exc_data]
    else:
        # LOG EXCEPTION HERE
        # with sentry_sdk.configure_scope() as scope:
        #     scope.user = {
        #         'id': request.user.id,
        #         'email': request.user.email,
        #     }
        sentry_sdk.capture_exception()
        logger.error(
            'ERROR: {}.{}'.format(type(exception).__module__, type(exception).__name__),
            exc_info=True,
            extra={'event': event, 'context': context},
        )
        if exc_data.__class__ in [list, tuple]:
            internal_non_field_errors = exc_data
        elif isinstance(exc_data, str):
            internal_non_field_errors = [exc_data]

    if 'errors' not in resp_data:
        resp_data['errors'] = {
            'non_field_errors': non_field_errors or ['Something went wrong. Please try later or contact admin.']
        }
    if internal_non_field_errors:
        resp_data['errors']['internal_non_field_errors'] = internal_non_field_errors
    resp_data['status_code'] = status_code
    return resp_data


# TODO: Write test
class LambdaDecorator(OLambdaDecorator):
    def __call__(self, event, context):
        self.event = event
        self.context = context
        return super().__call__(event, context)

    def before(self, event, context):
        response = load_json_body(lambda *args: args)(event, context)
        if type(response) == dict:  # not event, context
            raise ValidationError('Body parse error')
        if 'body' in event:
            event['body'] = underscoreize(event['body'] or {})
        return event, context

    @json_http_resp(cls=DjangoJSONEncoder)
    @cors_headers(origin=CORS_DOMAIN, credentials=True)
    @camelize_response
    def after(self, retval):
        return retval

    @json_http_resp(cls=DjangoJSONEncoder)
    @cors_headers(origin=CORS_DOMAIN, credentials=True)
    @camelize_response
    def on_exception(self, exception):
        return create_error_response(exception, self.event, self.context)



import os
import secrets
import urllib
from mimetypes import guess_extension
from urllib.parse import urlparse

try:
    from django.conf import settings as django_settings
except ImportError:
    pass


USER_AGENT = 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_1)' + \
    ' AppleWebKit/537.36 (KHTML, like Gecko) Chrome/39.0.2171.95 Safari/537.36'

DEFAULT_HEADERS = {
    'User-Agent': USER_AGENT,
}


def setting(name, *args):
    """
    key, fallback value
    """
    default = None
    default_defined = False
    if len(args):
        default = args[0]
        default_defined = True

    if not os.environ.get('IN_LAMBDA', False):
        # Use django settings for non lambda (for core deep)
        # DEEP_SERVERLESS is a prefix
        if default_defined:
            return getattr(django_settings, f'DEEP_SERVERLESS_{name}', default)
        return getattr(django_settings, f'DEEP_SERVERLESS_{name}')
    # In labmda use environment variables directly
    if default_defined:
        return os.environ.get(name, default)
    return os.environ[name]


def write_file(r, fp):
    for chunk in r.iter_content(chunk_size=1024):
        if chunk:
            fp.write(chunk)
    return fp


# TODO: same as get_filename_from_url
# This should also return filename without extension
def get_file_name_from_url(url):
    try:
        return urllib.parse.unquote(
            os.path.basename(
                urlparse(url).path
            ).split('.')[0]
        )
    except Exception:
        pass


def get_filename_from_url(url):
    return os.path.basename(urlparse(url).path)


def get_filename_from_base64(content):
    name = get_random_string(7)
    try:
        return name + guess_extension(content.split(';')[0].split(':')[1])
    except Exception:
        pass
    return name


def get_random_string(length, allowed_chars=(
    'abcdefghijklmnopqrstuvwxyz'
    'ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'
)):
    """
    SOURCE: DJANGO
    Return a securely generated random string.
    The bit length of the returned value can be calculated with the formula:
        log_2(len(allowed_chars)^length)
    For example, with default `allowed_chars` (26+26+10), this gives:
      * length: 12, bit length =~ 71 bits
      * length: 22, bit length =~ 131 bits
    """
    return ''.join(secrets.choice(allowed_chars) for i in range(length))



import os
from io import BytesIO, StringIO, UnsupportedOperation


class FileProxyMixin:
    """
    A mixin class used to forward file methods to an underlaying file
    object.  The internal file object has to be called "file"::

        class FileProxy(FileProxyMixin):
            def __init__(self, file):
                self.file = file
    """

    encoding = property(lambda self: self.file.encoding)
    fileno = property(lambda self: self.file.fileno)
    flush = property(lambda self: self.file.flush)
    isatty = property(lambda self: self.file.isatty)
    newlines = property(lambda self: self.file.newlines)
    read = property(lambda self: self.file.read)
    readinto = property(lambda self: self.file.readinto)
    readline = property(lambda self: self.file.readline)
    readlines = property(lambda self: self.file.readlines)
    seek = property(lambda self: self.file.seek)
    tell = property(lambda self: self.file.tell)
    truncate = property(lambda self: self.file.truncate)
    write = property(lambda self: self.file.write)
    writelines = property(lambda self: self.file.writelines)

    @property
    def closed(self):
        return not self.file or self.file.closed

    def readable(self):
        if self.closed:
            return False
        if hasattr(self.file, 'readable'):
            return self.file.readable()
        return True

    def writable(self):
        if self.closed:
            return False
        if hasattr(self.file, 'writable'):
            return self.file.writable()
        return 'w' in getattr(self.file, 'mode', '')

    def seekable(self):
        if self.closed:
            return False
        if hasattr(self.file, 'seekable'):
            return self.file.seekable()
        return True

    def __iter__(self):
        return iter(self.file)


class File(FileProxyMixin):
    DEFAULT_CHUNK_SIZE = 64 * 2 ** 10

    def __init__(self, file, name=None):
        self.file = file
        if name is None:
            name = getattr(file, 'name', None)
        self.name = name
        if hasattr(file, 'mode'):
            self.mode = file.mode

    def __str__(self):
        return self.name or ''

    def __repr__(self):
        return "<%s: %s>" % (self.__class__.__name__, self or "None")

    def __bool__(self):
        return bool(self.name)

    def __len__(self):
        return self.size

    def size(self):
        if hasattr(self.file, 'size'):
            return self.file.size
        if hasattr(self.file, 'name'):
            try:
                return os.path.getsize(self.file.name)
            except (OSError, TypeError):
                pass
        if hasattr(self.file, 'tell') and hasattr(self.file, 'seek'):
            pos = self.file.tell()
            self.file.seek(0, os.SEEK_END)
            size = self.file.tell()
            self.file.seek(pos)
            return size
        raise AttributeError("Unable to determine the file's size.")

    def chunks(self, chunk_size=None):
        """
        Read the file and yield chunks of ``chunk_size`` bytes (defaults to
        ``File.DEFAULT_CHUNK_SIZE``).
        """
        chunk_size = chunk_size or self.DEFAULT_CHUNK_SIZE
        try:
            self.seek(0)
        except (AttributeError, UnsupportedOperation):
            pass

        while True:
            data = self.read(chunk_size)
            if not data:
                break
            yield data

    def multiple_chunks(self, chunk_size=None):
        """
        Return ``True`` if you can expect multiple chunks.

        NB: If a particular file representation is in memory, subclasses should
        always return ``False`` -- there's no good reason to read from memory in
        chunks.
        """
        return self.size > (chunk_size or self.DEFAULT_CHUNK_SIZE)

    def __iter__(self):
        # Iterate over this file-like object by newlines
        buffer_ = None
        for chunk in self.chunks():
            for line in chunk.splitlines(True):
                if buffer_:
                    if endswith_cr(buffer_) and not equals_lf(line):
                        # Line split after a \r newline; yield buffer_.
                        yield buffer_
                        # Continue with line.
                    else:
                        # Line either split without a newline (line
                        # continues after buffer_) or with \r\n
                        # newline (line == b'\n').
                        line = buffer_ + line
                    # buffer_ handled, clear it.
                    buffer_ = None

                # If this is the end of a \n or \r\n line, yield.
                if endswith_lf(line):
                    yield line
                else:
                    buffer_ = line

        if buffer_ is not None:
            yield buffer_

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, tb):
        self.close()

    def open(self, mode=None):
        if not self.closed:
            self.seek(0)
        elif self.name and os.path.exists(self.name):
            self.file = open(self.name, mode or self.mode)
        else:
            raise ValueError("The file cannot be reopened.")
        return self

    def close(self):
        self.file.close()


class ContentFile(File):
    """
    A File-like object that takes just raw content, rather than an actual file.
    """
    def __init__(self, content, name=None):
        stream_class = StringIO if isinstance(content, str) else BytesIO
        super().__init__(stream_class(content), name=name)
        self.size = len(content)

    def __str__(self):
        return 'Raw content'

    def __bool__(self):
        return True

    def open(self, mode=None):
        self.seek(0)
        return self

    def close(self):
        pass

    def write(self, data):
        self.__dict__.pop('size', None)  # Clear the computed size.
        return self.file.write(data)


def endswith_cr(line):
    """Return True if line (a text or bytestring) ends with '\r'."""
    return line.endswith('\r' if isinstance(line, str) else b'\r')


def endswith_lf(line):
    """Return True if line (a text or bytestring) ends with '\n'."""
    return line.endswith('\n' if isinstance(line, str) else b'\n')


def equals_lf(line):
    """Return True if line (a text or bytestring) equals '\n'."""
    return line == ('\n' if isinstance(line, str) else b'\n')



import time
import uuid
import json

from pynamodb.models import Model
from pynamodb.attributes import (
    UnicodeAttribute,
    NumberAttribute,
    UnicodeSetAttribute,
    JSONAttribute as PydbJSONAttribute,
)

from ..utils.common import setting
from ..utils.django_utils.common import DjangoJSONEncoder
from ..utils.s3 import Storage


deep_media_storage = Storage(
    setting('MEDIA_BUCKET_NAME'),
    setting('MEDIA_BUCKET_ROOT', 'media'),
)


class JSONAttribute(PydbJSONAttribute):
    def serialize(self, value):
        """
        Serializes JSON to unicode
        """
        if value is None:
            return None
        encoded = json.dumps(value, cls=DjangoJSONEncoder)
        return encoded


class BaseModelMeta():
    if setting('IS_OFFLINE', False):
        host: str = 'http://localhost:8000'
        aws_access_key_id: str = 'DEFAULT_ACCESS_KEY'
        aws_secret_access_key: str = 'DEFAULT_SECRET'


class AsyncJob(Model):
    class Status():
        PENDING: str = 'pending'
        STARTED: str = 'started'
        SUCCESS: str = 'success'
        ERROR: str = 'error'  # expected
        FAILED: str = 'failed'  # unexpected

    class Type():
        SOURCE_EXTRACT: str = 'source_extract'

    class Meta(BaseModelMeta):
        table_name: str = setting('ASYNC_JOB_TABLE_NAME')

    uuid = UnicodeAttribute(hash_key=True, default=lambda: str(uuid.uuid4()))
    status = UnicodeAttribute(default=Status.PENDING)
    type = UnicodeAttribute()
    # Store key of the entities here
    entities = UnicodeSetAttribute()
    ttl = NumberAttribute(default=lambda: int(time.time()) + 86400)  # 1 day



import os
import sentry_sdk
from sentry_sdk.integrations.aws_lambda import AwsLambdaIntegration


# not using utils.common.setting instead of os.environ as this module shouldn't be used in deep core.
sentry_config = {
    'dsn': os.environ['SENTRY_DNS'],
    'environment': os.environ['STAGE'],
    'send_default_pii': True,
}

sentry_sdk.init(
    **sentry_config,
    integrations=[AwsLambdaIntegration()]
)



import uuid
import decimal
import json
import datetime


def is_aware(value):
    # https://github.com/django/django/blob/master/django/utils/timezone.py#L202
    return value.utcoffset() is not None


def _get_duration_components(duration):
    # https://github.com/django/django/blob/master/django/utils/duration.py#L4
    days = duration.days
    seconds = duration.seconds
    microseconds = duration.microseconds

    minutes = seconds // 60
    seconds = seconds % 60

    hours = minutes // 60
    minutes = minutes % 60

    return days, hours, minutes, seconds, microseconds


def duration_iso_string(duration):
    # https://github.com/django/django/blob/master/django/utils/duration.py#L31
    if duration < datetime.timedelta(0):
        sign = '-'
        duration *= -1
    else:
        sign = ''

    days, hours, minutes, seconds, microseconds = _get_duration_components(duration)
    ms = '.{:06d}'.format(microseconds) if microseconds else ""
    return '{}P{}DT{:02d}H{:02d}M{:02d}{}S'.format(sign, days, hours, minutes, seconds, ms)


class Promise:
    # https://github.com/django/django/blob/master/django/utils/functional.py#L68
    pass


class DjangoJSONEncoder(json.JSONEncoder):
    # https://github.com/django/django/blob/master/django/core/serializers/json.py#L77
    def default(self, o):
        if isinstance(o, datetime.datetime):
            r = o.isoformat()
            if o.microsecond:
                r = r[:23] + r[26:]
            if r.endswith('+00:00'):
                r = r[:-6] + 'Z'
            return r
        elif isinstance(o, datetime.date):
            return o.isoformat()
        elif isinstance(o, datetime.time):
            if is_aware(o):
                raise ValueError("JSON can't represent timezone-aware times.")
            r = o.isoformat()
            if o.microsecond:
                r = r[:12]
            return r
        elif isinstance(o, datetime.timedelta):
            return duration_iso_string(o)
        elif isinstance(o, (decimal.Decimal, uuid.UUID, Promise)):
            return str(o)
        else:
            return super().default(o)



import hashlib
import os
import sys

from pynamodb.models import Model
from pynamodb.attributes import (
    MapAttribute,
    UnicodeAttribute,
    NumberAttribute,
    UnicodeSetAttribute,
)

from ..utils.common import setting
from ..utils.django_utils.file import ContentFile

from .base import (
    deep_media_storage,

    BaseModelMeta,
    JSONAttribute,
)


class LeadExtract(MapAttribute):
    simplified_text = UnicodeAttribute(null=True)
    simplified_text_s3_key = UnicodeAttribute(null=True)
    word_count = NumberAttribute(null=True)
    page_count = NumberAttribute(null=True)
    file_size = NumberAttribute(null=True)


class Source(Model):
    class Status():
        PENDING: str = 'pending'
        STARTED: str = 'started'
        SUCCESS: str = 'success'
        ERROR: str = 'error'  # expected
        FAILED: str = 'failed'  # unexpected
        NOT_PROCESSED: str = 'not_processed'

    class Type():
        WEB: str = 'web'
        S3: str = 's3'

    # Make sure to update serverlesss.yml
    key = UnicodeAttribute(hash_key=True)  # Hash for url and `s3:<path>` for s3 objects
    # Make sure one of (url, s3_path) is defined
    url = UnicodeAttribute(null=True)
    s3_path = UnicodeAttribute(null=True)
    last_extracted_at = NumberAttribute(null=True)  # Epoch
    type = UnicodeAttribute()
    doc_type = UnicodeAttribute(null=True)
    extract = LeadExtract(default=dict)
    images = UnicodeSetAttribute(default=set())
    # Store extra information
    extra_meta = JSONAttribute(null=True)
    status = UnicodeAttribute(default=Status.PENDING)
    extract_status = UnicodeAttribute(default=Status.NOT_PROCESSED)
    thumbnail_status = UnicodeAttribute(default=Status.NOT_PROCESSED)

    class Meta(BaseModelMeta):
        table_name: str = setting('SOURCE_TABLE_NAME')

    @staticmethod
    def get_url_hash(url):
        return hashlib.sha224(url.encode()).hexdigest()

    @staticmethod
    def get_s3_key(s3_path):
        return f's3::{s3_path}'

    @property
    def usable_url(self):
        # Return url which can be used to access resource (web page, s3 file)
        if self.type == Source.Type.WEB:
            return self.url
        if self.type == Source.Type.S3:
            return deep_media_storage.url(self.s3_path)

    def get_file(self):
        if self.s3_path and self.type == Source.Type.S3:
            return deep_media_storage.get_file(self.s3_path)
        return None

    def upload_image(self, name, image):
        image_s3_path = deep_media_storage.upload(
            # This is from the-deep/server apps/lead/models.py:LeadPreviewImage:file
            os.path.join('lead-preview/from-lambda/', name),
            image,
        )
        self.images.add(image_s3_path)

    def set_simplified_text(self, text):
        if self.key is None:
            raise Exception('key needs to be defined')

        # Safe value (dynamodb item should not exceed 400K), TODO: Check for absolute
        if sys.getsizeof(text) < 400000:
            self.extract.simplified_text = text
        else:
            simplified_text_s3_key = deep_media_storage.upload(
                f'serverless/source-simplified-text/{self.key}.txt',
                ContentFile(text.encode('utf-8')),
            )
            self.extract.simplified_text_s3_key = simplified_text_s3_key

    def get_simplified_text(self):
        # NOTE: This may require KMS key access if files are encrypted in S3
        s3_key = self.extract.simplified_text_s3_key
        if s3_key:
            return deep_media_storage.get_file(s3_key).read().decode('utf-8')
        return self.extract.simplified_text

    def serialize(self):
        # TODO: Use serializer utils
        return {
            'key': self.key,
            'url': self.url,
            'type': self.type,
            'status': self.status,
            'last_extracted_at': self.last_extracted_at,
            'extra_meta': self.extra_meta,
        }



import time

from .common import get_file_name_from_url
from .validators import validator
from .camelize import camelize


class Timeout:
    def __init__(self, timeout=30):
        self.timeout = timeout

    def __call__(self, func_to_be_tracked):
        def wrapper(*args, **kwargs):
            start = time.time()
            ret = func_to_be_tracked(*args, **kwargs)
            total_time = time.time() - start
            assert self.timeout >= total_time, f'Total time should be <= 30 instead of {total_time}'
            return ret
        wrapper.__name__ = func_to_be_tracked.__name__
        wrapper.__module__ = func_to_be_tracked.__module__
        return wrapper


def test_get_file_name_from_url():
    url = 'https://reliefweb.int/sites/reliefweb.int/files/resources/ReliefWeb%20Vision%20and%20Strategy.pdf'
    filename = get_file_name_from_url(url)
    assert 'ReliefWeb Vision and Strategy' == filename


def test_validator():
    valid_urls = (
        'https://reliefweb.int/sites/reliefweb.int/files/resources/ReliefWeb%20Vision%20and%20Strategy.pdf',
    )
    invalid_urls = (
        'https://reliefweb.int/sites/reliefweb.int/files/resources/ReliefWeb Vision and Strategy.pdf',
    )
    for url in valid_urls:
        assert validator.is_url(url) is True
    for url in invalid_urls:
        assert validator.is_url(url) is False


def test_camelize():
    res = camelize({
        'test_1': 'value 1',
        'test_2': 'value 1',
        'nested_test': {
            'nested_key_one': 'value',
            'nested_key_2': 'value',
            'nested_key_three': 'value',
        },
    })
    assert res == {
        'test1': 'value 1',
        'test2': 'value 1',
        'nestedTest': {
            'nestedKeyOne': 'value',
            'nestedKey2': 'value',
            'nestedKeyThree': 'value',
        },
    }



from setuptools import setup, find_namespace_packages


setup(
    name='deep-serverless',
    version='1.0.4',
    description='Sharable modules from Deep serverlss',
    long_description_content_type='text/markdown',
    long_description=open('./README.md').read(),
    license="MIT",

    author='TC Dev',
    author_email='dev@togglecorp.com',
    url='https://github.com/the-deep/serverless',

    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Topic :: Utilities",
        "Operating System :: OS Independent",
    ],

    packages=find_namespace_packages(),
    # packages=find_packages(exclude=("tests", "tests.*")),
    install_requires=[
        'boto3==1.9.88',
        'pynamodb==4.3.2',
        'requests==2.21.0',
    ],
    zip_safe=False,
)



import re


class ValidationError(Exception):
    pass


class Validator():
    VALID_URL_REGEX = re.compile(
        # http:// or https://
        r'^(?:http)s?://'
        # domain...
        r'(?:(?:[A-Z0-9]'
        r'(?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+'
        r'(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|'
        # localhost...
        r'localhost|'
        # ...or ip
        r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})'
        # optional port
        r'(?::\d+)?'
        r'(?:/?|[/?]\S+)$',
        re.IGNORECASE
    )

    @classmethod
    def is_url(cls, url: str) -> bool:
        if cls.VALID_URL_REGEX.match(url):
            return True
        return False


validator = Validator()



# SOURCE: https://github.com/django/django/blob/master/django/core/files/storage.py
# SOURCE: https://github.com/jschneier/django-storages/blob/master/storages/backends/s3boto3.py
# NOTE: Only copied used part.

import io
import os
import boto3
import posixpath
import mimetypes
from gzip import GzipFile
from tempfile import SpooledTemporaryFile
from botocore.exceptions import ClientError

from .common import get_random_string


def force_bytes(s):
    if isinstance(s, bytes):
        return s
    if isinstance(s, memoryview):
        return bytes(s)
    return str(s).encode('utf-8', 'strict')


def safe_join(base, *paths):
    base_path = base
    base_path = base_path.rstrip('/')
    paths = [p for p in paths]
    final_path = base_path + '/'
    for path in paths:
        _final_path = posixpath.normpath(posixpath.join(final_path, path))
        # posixpath.normpath() strips the trailing /. Add it back.
        if path.endswith('/') or _final_path + '/' == final_path:
            _final_path += '/'
        final_path = _final_path
    if final_path == base_path:
        final_path += '/'
    base_path_len = len(base_path)
    if (not final_path.startswith(base_path) or final_path[base_path_len] != '/'):
        raise ValueError('the joined path is located outside of the base path component')
    return final_path.lstrip('/')


class Storage():
    DEFAULT_CONTENT_TYPE = 'application/octet-stream'
    DEFAULT_QUERYSTRING_EXPIRE = 3600  # 1 hour

    QUERYSTRING_AUTH = True
    GZIP_CONTENT_TYPES = [
        'text/javascript', 'application/javascript', 'application/x-javascript',
        'image/svg+xml', 'application/json', 'text/css',
    ]

    def __init__(self, bucket_name, base_path):
        self.bucket_name = bucket_name
        self.base_path = base_path
        self._file = None

    @property
    def connection(self):
        connection = getattr(self, '_connection', None)
        if connection is None:
            session = boto3.session.Session()
            self._connection = session.resource('s3')
        return self._connection

    @property
    def bucket(self):
        bucket = getattr(self, '_bucket', None)
        if bucket is None:
            self._bucket = self.connection.Bucket(self.bucket_name)
        return self._bucket

    def exists(self, name):
        name = self._normalize_name(name)
        try:
            self.connection.meta.client.head_object(Bucket=self.bucket_name, Key=name)
            return True
        except ClientError:
            return False

    def get_alternative_name(self, file_root, file_ext):
        return '%s_%s%s' % (file_root, get_random_string(7), file_ext)

    # NOTE: max_length is defined by the max_length for django FileField default configuration. (Don't change this here)
    def get_available_name(self, name, max_length=100):
        dir_name, file_name = os.path.split(name)
        file_root, file_ext = os.path.splitext(file_name)
        while self.exists(name) or (max_length and len(name) > max_length):
            name = os.path.join(dir_name, self.get_alternative_name(file_root, file_ext))
            if max_length is None:
                continue
            truncation = len(name) - max_length
            if truncation > 0:
                file_root = file_root[:-truncation]
                if not file_root:
                    raise Exception(f"Can't generate filename for '{name}' due to s3 key size limit.")
                name = os.path.join(dir_name, self.get_alternative_name(file_root, file_ext))
        return name

    def _compress_content(self, content):
        """Gzip a given string content."""
        content.seek(0)
        zbuf = io.BytesIO()
        zfile = GzipFile(mode='wb', fileobj=zbuf, mtime=0.0)
        try:
            zfile.write(force_bytes(content.read()))
        finally:
            zfile.close()
        zbuf.seek(0)
        return zbuf

    def _clean_name(self, name):
        clean_name = posixpath.normpath(name).replace('\\', '/')
        if name.endswith('/') and not clean_name.endswith('/'):
            clean_name += '/'
        return clean_name

    def _normalize_name(self, name):
        try:
            return safe_join(self.base_path, name)
        except ValueError:
            raise Exception("Attempted access to '%s' denied." % name)

    def _get_write_parameters(self, name, content=None):
        params = {}
        _type, encoding = mimetypes.guess_type(name)
        content_type = getattr(content, 'content_type', None)
        content_type = content_type or _type or self.DEFAULT_CONTENT_TYPE
        params['ContentType'] = content_type
        if encoding:
            params['ContentEncoding'] = encoding
        return params

    def upload(self, name, content, gzip=True):
        if not name:
            raise Exception('Name is required')
        cleaned_name = self.get_available_name(self._clean_name(name))
        name = self._normalize_name(cleaned_name)
        params = self._get_write_parameters(name, content)
        if (
            gzip and
            params['ContentType'] in self.GZIP_CONTENT_TYPES and
            'ContentEncoding' not in params
        ):
            content = self._compress_content(content)
            params['ContentEncoding'] = 'gzip'
        content.seek(0, os.SEEK_SET)
        obj = self.bucket.Object(name)
        obj.upload_fileobj(content, ExtraArgs=params)
        return cleaned_name

    def url(self, name, parameters=None, expire=None, http_method=None):
        name = self._normalize_name(self._clean_name(name))
        if expire is None:
            expire = self.DEFAULT_QUERYSTRING_EXPIRE
        params = parameters.copy() if parameters else {}
        params['Bucket'] = self.bucket.name
        params['Key'] = name
        url = self.bucket.meta.client.generate_presigned_url(
            'get_object', Params=params, ExpiresIn=expire, HttpMethod=http_method
        )
        if self.QUERYSTRING_AUTH:
            return url
        return self._strip_signing_parameters(url)

    def get_file(self, name):
        name = self._normalize_name(self._clean_name(name))
        file = SpooledTemporaryFile(
            suffix=".S3Boto3StorageFile",
            dir='/tmp/'
        )
        obj = self.bucket.Object(name)
        obj.download_fileobj(file)
        file.seek(0)
        return file
